import glob
import os
from subprocess import Popen, PIPE
from porkchop.plugin import PorkchopPlugin

class RunitPlugin(PorkchopPlugin):
  def service_status(self, service):
    return Popen('/usr/bin/sv status %s' % service, stdout=PIPE,
      shell=True).communicate()[0].strip()

  def get_data(self):
    d1 = {}

    service_dirs = glob.glob('/etc/sv/*')
    for dir in service_dirs:
      service = os.path.basename(dir)
      d1[service] = self.service_status(service)

    return d1



from collections import defaultdict
import psycopg2
import psycopg2.extras

from porkchop.plugin import PorkchopPlugin

def exc(conn, query, col_key = None, col_val = None):
  data = {}
  cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
  cursor.execute(query)

  if col_key and col_val:
    for row in cursor.fetchall():
      data[row[col_key]] = row[col_val]
  else:
    data = cursor.fetchall()

  return data

class PostgresqlPlugin(PorkchopPlugin):
  def _connect(self, dbname = None):
    dbname = dbname if dbname else 'postgres'
    try:
      conn_string = "host='%s' dbname='%s' user='%s' password='%s'" %\
        (self.config['postgresql']['host'],
        dbname,
        self.config['postgresql']['user'],
        self.config['postgresql']['password'])

      conn = psycopg2.connect(conn_string)
      conn.set_isolation_level(0) # don't need transactions
    except:
      return None

    return conn

  def get_data(self):
    data = self.gendict()
    conn = self._connect()

    datname_list_query = "SELECT datname FROM pg_database \
                         WHERE datallowconn AND NOT datistemplate \
                         AND NOT datname='postgres' ORDER BY 1 LIMIT 10"
    buffercache_query = 'SELECT sum(blks_read) \
                        AS blks_read,sum(blks_hit) AS blks_hit \
                        FROM pg_stat_database'
    all_conn_count_query = "SELECT tmp.state,COALESCE(count,0) FROM \
                           (VALUES ('active'),('waiting'),('idle'),\
                           ('idletransaction'),('unknown')) AS tmp(state) \
                           LEFT JOIN \
                           (SELECT CASE WHEN waiting THEN 'waiting' \
                             WHEN current_query='<IDLE>' THEN 'idle' \
                             WHEN current_query='<IDLE> in transaction' \
                             THEN 'idletransaction' \
                             WHEN current_query='<insufficient privilege>' \
                             THEN 'unknown' ELSE 'active' END AS state, \
                             count(*) AS count \
                             FROM pg_stat_activity \
                             WHERE procpid != pg_backend_pid() \
                             GROUP BY CASE WHEN waiting THEN 'waiting' \
                             WHEN current_query='<IDLE>' THEN 'idle' \
                             WHEN current_query='<IDLE> in transaction' \
                             THEN 'idletransaction' \
                             WHEN current_query='<insufficient privilege>' \
                             THEN 'unknown' ELSE 'active' END \
                           ) AS tmp2 \
                           ON tmp.state=tmp2.state ORDER BY 1"
    db_conn_count_query = 'SELECT pg_database.datname, \
                          COALESCE(count,0) AS count FROM pg_database \
                          LEFT JOIN \
                          (SELECT datname,count(*) FROM pg_stat_activity \
                          WHERE procpid != pg_backend_pid() \
                          GROUP BY datname) AS tmp \
                          ON pg_database.datname=tmp.datname \
                          WHERE datallowconn ORDER BY 1'
    user_conn_count_query = 'SELECT usename,count(*) \
                            FROM pg_stat_activity \
                            WHERE procpid != pg_backend_pid() \
                            GROUP BY usename ORDER BY 1'
    locks_query = 'SELECT lower(mode) AS mode,count(*) \
                  FROM pg_locks WHERE database \
                  IS NOT NULL GROUP BY mode ORDER BY 1'
    db_size_query = 'SELECT datname,pg_database_size(oid) \
                     FROM pg_database ORDER BY 1'
    query_time_query = "SELECT 'query' AS type, COALESCE(max(extract(epoch \
                       FROM CURRENT_TIMESTAMP-query_start)),0) \
                       FROM pg_stat_activity WHERE current_query \
                       NOT LIKE '<IDLE%' UNION ALL \
                       SELECT 'transaction', \
                       COALESCE(max(extract(epoch \
                       FROM CURRENT_TIMESTAMP-xact_start)),0) \
                       FROM pg_stat_activity WHERE 1=1"
    idle_xact_time_query = "SELECT datname, \
                            max(COALESCE(ROUND(EXTRACT(epoch \
                            FROM now()-query_start)),0)) FROM pg_stat_activity \
                            WHERE current_query = '<IDLE> in transaction' \
                            GROUP BY 1"
    xact_count_query = "SELECT 'commit' AS type, \
                       sum(pg_stat_get_db_xact_commit(oid)) FROM pg_database \
                       UNION ALL SELECT 'rollback', \
                       sum(pg_stat_get_db_xact_rollback(oid)) FROM pg_database"
    scan_type_query = 'SELECT COALESCE(sum(seq_scan),0) AS sequential, \
                      COALESCE(sum(idx_scan),0) AS index \
                      FROM pg_stat_user_tables'
    xlog_query = "SELECT count(*) AS segments FROM pg_ls_dir('pg_xlog') t(fn) \
                 WHERE fn ~ '^[0-9A-Z]{24}\$'"
    tuple_access_query = "SELECT COALESCE(sum(seq_tup_read),0) AS seqread, \
                         COALESCE(sum(idx_tup_fetch),0) AS idxfetch, \
                         COALESCE(sum(n_tup_ins),0) AS inserted, \
                         COALESCE(sum(n_tup_upd),0) AS updated, \
                         COALESCE(sum(n_tup_del),0) AS deleted, \
                         COALESCE(sum(n_tup_hot_upd),0) AS hotupdated \
                         FROM pg_stat_user_tables"

    data['query_time'] = exc(conn, query_time_query, 'type', 'coalesce')
    datnames = exc(conn, datname_list_query)[0]
    data['datnames'] = ', '.join(datnames)

    row = exc(conn, 'SELECT * from pg_stat_bgwriter')[0]
    for key in row.keys():
      data['bgwriter'][key] = row[key]

    row = exc(conn, buffercache_query)[0]
    for key in row.keys():
      data['buffercache'][key] = row[key]

    data['connections'] = {
      'state': exc(conn, all_conn_count_query, 'state', 'coalesce'),
      'database': exc(conn, db_conn_count_query, 'datname', 'count'),
      'user': exc(conn, user_conn_count_query, 'usename', 'count')
    }

    data['size'] = exc(conn, db_size_query, 'datname', 'pg_database_size')
    data['locks'] = exc(conn, locks_query, 'mode', 'count')
    data['idle_xact_time'] = exc(conn, idle_xact_time_query, 'datname', 'max')
    data['xactn_count'] = exc(conn, xact_count_query, 'type', 'sum')

    row = exc(conn, xlog_query)[0]
    for key in row.keys():
      data['xlog'][key] = row[key]

    for db in datnames:
      conn2 = self._connect(db)
      row = exc(conn2, tuple_access_query)[0]
      for key in row.keys():
        data['tuple_access'][db][key] = row[key]

      row = exc(conn2, scan_type_query)[0]
      for key in row.keys():
        data['scans'][db][key] = row[key]

    try:
      slony_db = self.config['postgresql']['slony_db']
      slony_node_string = self.config['postgresql']['slony_node_string']
      slony_table = self.config['postgresql']['slony_table']

      replication_lag_query = "SELECT SUBSTRING(sl.no_comment \
                              FROM '%s') AS NODE, \
                              st.st_lag_num_events AS lag_events \
                              FROM %s.sl_status AS st, %s.sl_node AS sl \
                              WHERE sl.no_id = st.st_received" % \
                              (slony_node_string, slony_table, slony_table)

      conn2 = self._connect(slony_db)
      data['slony_replication'] = exc(conn2, replication_lag_query,
                                      'node',
                                      'lag_events')
    except:
      pass

    return data



from subprocess import Popen, PIPE
import time

from porkchop.plugin import PorkchopPlugin

def varnish_stats():
  data = {}
  cmd = 'varnishstat -1'
  output = Popen([cmd], stdout=PIPE, shell=True).\
                 communicate()[0].splitlines()

  for line in output:
    fields = line.split()
    metric = fields[0]
    longterm_rate = fields[2] if fields[2] != '.' else 0
    data[metric] = {
      'counter': fields[1],
      'longterm_per_second': longterm_rate
    }

  return data

class VarnishPlugin(PorkchopPlugin):
  def get_data(self):
    if not self.__class__._cache:
      prev = varnish_stats()
      ival = 1
      time.sleep(ival)
    else:
      prev = self.__class__._cache
      ival = int(time.time() - self.__class__._lastrefresh)

    data = self.__class__._cache = cur = varnish_stats()

    for metric in data.keys():
      if data[metric]['longterm_per_second'] > 0:
        data[metric]['shortterm_per_second'] =\
          self.rateof(prev[metric]['counter'],
                      cur[metric]['counter'], ival)

    return data



import glob
import psycopg2
import psycopg2.extras
import re

from porkchop.plugin import PorkchopPlugin

def exc(conn, query, col_key = None, col_val = None):
  data = {}
  cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
  cursor.execute(query)

  if col_key and col_val:
    for row in cursor.fetchall():
      data[row[col_key]] = row[col_val]
  else:
    data = cursor.fetchall()

  return data

class PgbouncerPlugin(PorkchopPlugin):
  def _connect(self, **kwargs):
    host = kwargs.get('host', 'localhost')
    port = kwargs.get('port', '6432')
    user = kwargs.get('user', 'psql')
    password = kwargs.get('password', '')

    try:
      conn_string = "host='%s' port='%s' dbname='pgbouncer' user='%s'\
                    password='%s'" % (host, port, user, password)

      conn = psycopg2.connect(conn_string)
      conn.set_isolation_level(0) # don't need transactions
    except:
      return None

    return conn

  def get_data(self):
    data = self.gendict()

    instances = self._get_instances()

    for instance in instances.keys():
      # haven't decided how to best handle auth
      try:
        conn = self._connect(host=instances[instance]['addr'],
                             port=instances[instance]['port'])
        data[instance] = self._instance_stats(conn)
      except:
        pass

    return data

  def _get_instances(self):
    configs = glob.glob('/etc/pgbouncer/*.ini')
    r1 = re.compile('^listen_(addr|port)\s+=\s+(.*)$')
    instances = self.gendict()

    for file in configs:
      inst = file.split('/')[3].split('.')[0]

      with open(file, 'r') as f:
        for line in f:
          match = r1.match(line)
          if match:
            instances[inst].update({match.group(1): match.group(2)})

    return instances

  def _instance_stats(self, conn):
    data = self.gendict()

    for row in exc(conn, 'show stats'):
      for col in [key for key in row.keys() if key != 'database']:
        data[row['database']][col] = row[col]

    res = exc(conn, 'show servers')
    for x in xrange(len(res)):
      for col in [key for key in res[x].keys()]:
        data['server%d' % x][col] = res[x][col]

    res = exc(conn, 'show clients')
    for x in xrange(len(res)):
      for col in [key for key in res[x].keys()]:
        data['client%d' % x][col] = res[x][col]

    for row in exc(conn, 'show pools'):
      for col in [key for key in row.keys() if key != 'database']:
        data[row['database']][col] = row[col]

    data['lists'] = exc(conn, 'show lists', 'list', 'items')

    for row in exc(conn, 'show databases'):
      for col in [key for key in row.keys() if key != 'name']:
        data[row['database']][col] = row[col]

    for row in exc(conn, 'show fds'):
      for col in [key for key in row.keys() if key != 'fd']:
        data['fd%d' % row['fd']][col] = row[col]

    return data



import re
from subprocess import Popen, PIPE

from porkchop.plugin import PorkchopPlugin

def df(cmd):
  data = {}
  r1 = re.compile('^/(dev/|)')

  output = Popen([cmd], stdout=PIPE, shell=True).communicate()[0].splitlines()

  output.pop(0)
  for line in output:
    fields = line.split()
    key = r1.sub('', fields[0]).replace('/', '_')
    data[key] = {
      'total': fields[1],
      'used': fields[2],
      'available': fields[3],
      'percent_used': fields[4].replace('-', '0').replace('%', ''),
      'mountpoint': fields[5]
    }

  return data

class DfPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}
    cmd = 'df -l'
    excludefs = [
      'debugfs',
      'devtmpfs',
      'ecryptfs',
      'iso9660',
      'none',
      'ramfs',
      'romfs',
      'squashfs',
      'simfs'
      'udf',
      'unknown',
      'tmpfs'
    ]

    blocks_cmd = '%s -x %s' % (cmd, ' -x '.join(excludefs))
    inodes_cmd = '%s -i -x %s' % (cmd, ' -x '.join(excludefs))

    blocks = df(blocks_cmd)
    for key in blocks:
      data.setdefault(key, {})
      data[key]['mountpoint'] = blocks[key]['mountpoint']
      del blocks[key]['mountpoint']
      data[key]['blocks'] = blocks[key]

    inodes = df(inodes_cmd)
    for key in inodes:
      data.setdefault(key, {})
      del inodes[key]['mountpoint']
      data[key]['inodes'] = inodes[key]

    return data



import logging
import socket
from optparse import OptionParser

from porkchop.plugin import PorkchopPluginHandler

def coerce_number(s):
  try:
    return int(s)
  except:
    return float(s)

def get_logger(name, level=logging.INFO):
  logger = logging.getLogger(name)
  logger.setLevel(logging.DEBUG)
  ch = logging.StreamHandler()
  ch.setLevel(level)
  formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
  ch.setFormatter(formatter)
  logger.addHandler(ch)

  return logger

def main():
  from server import GetHandler
  from porkchop.server import GetHandler, ThreadedHTTPServer
  config_dir = '/etc/porkchop.d'
  plugin_dir = '/usr/share/porkchop/plugins'
  listen_address = ''
  listen_port = 5000

  parser = OptionParser()
  parser.add_option('-c', dest='config_dir',
                    default=config_dir,
                    help='Load configs from DIR (default: %s)' % config_dir,
                    metavar='DIR')
  parser.add_option('-d', dest='plugin_dir',
                    default=plugin_dir,
                    help='Load plugins from DIR (default: %s)' % plugin_dir,
                    metavar='DIR')
  parser.add_option('-s', dest='listen_address',
                    default=listen_address,
                    help='Bind to ADDRESS', metavar='ADDRESS')
  parser.add_option('-p', type="int", dest='listen_port',
                    default=listen_port,
                    help='Bind to PORT (default: %d)' % listen_port,
                    metavar='PORT')

  (options, args) = parser.parse_args()

  PorkchopPluginHandler(options.config_dir, options.plugin_dir)
  server = ThreadedHTTPServer((options.listen_address, options.listen_port),
                              GetHandler)
  server.serve_forever()

def collector():
  import requests
  import sys
  import time

  from porkchop.backend import Carbon

  carbon_host = 'localhost'
  carbon_port = 2004
  data = {}
  porkchop_url = 'http://localhost:5000/'

  interval = 60
  prefix = 'porkchop.%s' % socket.gethostname().split('.')[0].replace('.','_')

  parser = OptionParser()
  parser.add_option('--carbon-host', dest='carbon_host',
                    default=carbon_host,
                    help='Connect to carbon on HOST (default: %s)' % carbon_host,
                    metavar='HOST')
  parser.add_option('--carbon-port', type='int', dest='carbon_port',
                    default=carbon_port,
                    help='Connect to carbon on PORT (default: %d)' % carbon_port,
                    metavar='PORT')
  parser.add_option('--porkchop-url', dest='porkchop_url',
                    default=porkchop_url,
                    help='Connect to porkchop on URL (default: %s)' % porkchop_url,
                    metavar='URL')
  parser.add_option('-i', type='int', dest='interval',
                    default=interval,
                    help='Fetch data at INTERVAL (default: %d)' % interval,
                    metavar='INTERVAL')
  parser.add_option('-n', dest='noop',
                    default=False,
                    help='Don\'t actually send to graphite',
                    action='store_true')
  parser.add_option('-P', dest='prefix',
                    default=prefix,
                    help='Graphite prefix (default: %s)' % prefix)
  parser.add_option('-v', dest='verbose',
                    default=False,
                    help='Verbose logging',
                    action='store_true')


  (options, args) = parser.parse_args()

  if options.verbose:
    logger = get_logger('porkchop-collector', logging.DEBUG)
  else:
    logger = get_logger('porkchop-collector')

  if not options.noop:
    carbon = Carbon(options.carbon_host, options.carbon_port, logger)

  while True:
    now = int(time.time())
    try:
      logger.debug('Fetching porkchop data from %s', options.porkchop_url)
      r = requests.get(options.porkchop_url, headers={'x-porkchop-refresh': 'true'})
      r.raise_for_status()
    except:
      logger.error('Got bad response code from porkchop: %s', sys.exc_info()[1])

    lines = []
    for line in r.content.strip('\n').splitlines():
      (key, val) = line.lstrip('/').split(' ', 1)
      key = '.'.join([options.prefix, key.replace('/', '.')])
      data.setdefault(key, [])

      try:
        data[key].append((now, coerce_number(val)))

        for met in data.keys():
          for datapoint in data[met]:
            logger.debug('Sending: %s %s %s', met, datapoint[0], datapoint[1])

        if not options.noop:
          carbon.data = data
          carbon.send()
      except:
        pass

    sleep_time = options.interval - (int(time.time()) - now)
    if sleep_time > 0:
      logger.info('Sleeping for %d seconds', sleep_time)
      time.sleep(sleep_time)



import socket
import struct
import time
import cPickle

class Carbon(object):
  def __init__(self, host, port, logger):
    self.data = {}
    self.host = host
    self.port = port
    self.logger = logger
    try:
      self.sock = self._connect()
    except socket.error:
      self.logger.fatal('Unable to connect to carbon.')

  def _connect(self, waittime=5):
    self.logger.info('Connecting to carbon on %s:%d', self.host, self.port)
    try:
      sock = socket.socket()
      sock.connect((self.host, self.port))
    except socket.error:
      self.logger.info('Unable to connect to carbon, retrying in %d seconds', waittime)
      time.sleep(waittime)
      self._connect(waittime + 5)

    return sock

  def _send(self, data):
    try:
      self.sock.sendall(self._serialize(data.items()))
    except socket.error:
      raise

  def _serialize(self, data):
    serialized = cPickle.dumps(data, protocol=-1)
    prefix = struct.pack('!L', len(serialized))
    return prefix + serialized

  def send(self):
    """ self.data format: {metric_name: [(t1, val1), (t2, val2)]} """
    buf_sz = 500
    to_send = {}

    for mn in self.data.keys():
      while len(self.data[mn]) > 0:
        l = len(to_send)
        if l < buf_sz:
          to_send.setdefault(mn, [])
          to_send[mn].append(self.data[mn].pop())
        else:
          try:
            self._send(to_send)
            to_send = {}
            to_send.setdefault(mn, [])
            to_send[mn].append(self.data[mn].pop())
          except socket.error:
            self.logger.error('Error sending to carbon, trying to reconnect.')
            self.sock = self._connect()

            # we failed to send, so put it back in the stack and try later
            for ent in to_send:
              self.data[ent[0]].append(ent[1])

    try:
      self._send(to_send)
    except socket.error:
      self.logger.error('Error sending to carbon, trying to reconnect.')
      self.sock = self._connect()

      # we failed to send, so put it back in the stack and try later
      for ent in to_send:
        self.data[ent[0]].append(ent[1])



import time

from porkchop.plugin import PorkchopPlugin

class NetprotoPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}

    if not self.__class__._cache:
      prev = self._read_info()
      delta = 1
      time.sleep(delta)
    else:
      prev = self.__class__._cache
      delta = int(time.time() - self.__class__._lastrefresh)

    self.__class__._cache = cur = self._read_info()

    for proto in cur.keys():
      data.setdefault(proto, {})
      for stat in cur[proto].keys():
        data[proto][stat] = self.rateof(prev[proto][stat],
                                        cur[proto][stat], delta)

    return data

  def _read_info(self):
    data = self.gendict()
    with open('/proc/net/snmp', 'r') as f:
      for line in f:
        fields = line.split()
        proto = fields[0].lower().rstrip(':')

        if not proto in data.keys():
          keys = [fld.lower() for fld in fields[1:]]
          data[proto] = {}
        else:
          data[proto] = dict(zip(keys, tuple(fields[1:])))

    return data



from collections import defaultdict
import time

from porkchop.plugin import PorkchopPlugin

class NetworkinterfacePlugin(PorkchopPlugin):
  def get_data(self):
    data = self.gendict()

    if not self.__class__._cache:
      prev = self._read_info()
      delta = 1
      time.sleep(delta)
    else:
      prev = self.__class__._cache
      delta = int(time.time() - self.__class__._lastrefresh)

    self.__class__._cache = cur = self._read_info()

    for iface in cur.keys():
      for typ in cur[iface].keys():
        for stat in cur[iface][typ].keys():
          label = '%s_per_second' % stat
          data[iface][typ][label] = self.rateof(prev[iface][typ][stat],
                                                cur[iface][typ][stat],
                                                delta)

    return data

  def _read_info(self):
    data = self.gendict()

    with open('/proc/net/dev', 'r') as f:
      output = f.readlines()

    output.pop(0)

    keys = output[0].replace('|',' ').split()[1:]
    recv_keys = tuple(keys[0:8])
    xmit_keys = tuple(keys[8:])

    for line in output[1:]:
      (iface, fields) = line.strip().split(':')
      fields = fields.split()
      recv_tup = tuple(fields[0:8])
      xmit_tup = tuple(fields[8:])
      data[iface]['recv'] = dict(zip(recv_keys, recv_tup))
      data[iface]['xmit'] = dict(zip(xmit_keys, xmit_tup))

    return data



import re

from porkchop.plugin import PorkchopPlugin

class CpuinfoPlugin(PorkchopPlugin):
  def get_data(self):
    d1 = {}
    r1 = re.compile('(\w+)\s+:\s+(\w+)')
    with open('/proc/cpuinfo', 'r') as f:
      for line in f:
        match = r1.match(line)
        if match:
          k = match.group(1)
          v = match.group(2)
        if k == 'processor':
          proc = v
          d1['processor%s' % proc] = {}
        else:
          d1['processor%s' % proc].update({k: v})

    return d1



import re
import socket

from porkchop.plugin import PorkchopPlugin

HAPROXY_SOCKET = '/var/run/haproxy.socket'

class HAProxy(object):
  @classmethod
  def col2stat(self, idx):
    try:
      ret = {
        0: 'vip_name', 1: 'backend_name', 2: 'queue_cur',
        3: 'queue_max', 4: 'sessions_cur', 5: 'sessions_max',
        6: 'sessions_limit', 7: 'sessions_total', 8: 'bytes_in',
        9: 'bytes_out', 10: 'denied_req', 11: 'denied_resp',
        12: 'errors_req', 13: 'errors_conn', 14: 'errors_resp',
        15: 'warnings_retr', 16: 'warnings_redis', #17: 'server_status',
        18: 'server_weight', 19: 'server_active', 20: 'server_backup',
        21: 'server_check', 22: 'server_down', 23: 'server_lastchange',
        24: 'server_downtime', 25: 'queue_limit', 29: 'server_throttle',
        30: 'sessions_lbtotal', 32: 'session_type', 33: 'session_rate_cur',
        34: 'session_rate_limit', 35: 'session_rate_max'
      }[idx]
    except KeyError:
      ret = None

    return ret

  @classmethod
  def stats(self, sock_path):
    try:
      sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      sock.connect(sock_path)
      sock_data = []
      data = ''

      sock.send('show stat\n')

      while True:
        data = sock.recv(1024)
        if not data: break
        sock_data.append(data)

      sock.close()
    except socket.error:
      return []

    return ''.join(sock_data).strip('\n').splitlines()

class HaproxyPlugin(PorkchopPlugin):
  def get_data(self):
    all_stats = []
    d1 = {}
    r1 = re.compile('^(#|$)')

    for line in HAProxy.stats(HAPROXY_SOCKET):
      if r1.match(line): continue

      lary = line.split(',')
      stat = {}

      for i in xrange(0,len(lary)-1):
        if not lary[i]: lary[i] = None
        stat.update({HAProxy.col2stat(i): lary[i]})

      try:
        del stat[None]
      except:
        pass

      all_stats.append(stat)

    vip = {}
    d1['vip'] = {}
    for x in all_stats:
      vip.setdefault(x['vip_name'], {})
      vip[x['vip_name']].setdefault('backend', {})
      vip[x['vip_name']]['backend'].update({x['backend_name']: {}})


      for key, val in x.iteritems():
        r2 = re.compile('_name')
        vip[x['vip_name']]['backend'][x['backend_name']].update({key: val})

      d1['vip'].update(vip)

    return d1



import ConfigParser

class PorkchopUtil(object):
  @classmethod
  def parse_config(self, path):
    config = {}
    cp = ConfigParser.ConfigParser()
    cp.read(path)

    for s in cp.sections():
      config.setdefault(s, {})
      for o in cp.options(s):
        config[s][o] = cp.get(s, o)

    return config



import os
import re
import time

from porkchop.plugin import PorkchopPlugin

def fmt(f):
  return '%.2f' % f


def read_info():
  data = {}

  stat_fields = [
    'rd',
    'rd_m',
    'rd_s',
    'rd_t',
    'wr',
    'wr_m',
    'wr_s',
    'wr_t',
    'running',
    'use',
    'aveq'
  ]

  r1 = re.compile('[a-z]d[a-z]$')
  with open('/proc/diskstats', 'r') as f:
    for line in f:
      fields = line.strip().split()
      fields.pop(0); fields.pop(0)
      device = fields[0]
      fields.pop(0)

      if not r1.match(device): continue

      data.setdefault(device, {})

      for pos in xrange(len(fields)):
        data[device][stat_fields[pos]] = float(fields[pos])

  return data

class DiskstatsPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}

    if not self.__class__._cache:
      prev = read_info()
      delta = 1
      time.sleep(delta)
    else:
      prev = self.__class__._cache
      delta = int(time.time() - self.__class__._lastrefresh)

    self.__class__._cache = cur = read_info()

    for key in prev.keys():
      try:
        avg_req_sz = ((cur[key]['rd_s'] - prev[key]['rd_s']) +\
                      (cur[key]['wr_s'] - prev[key]['wr_s'])) /\
                     ((cur[key]['rd'] + cur[key]['wr']) -\
                      (prev[key]['rd'] + prev[key]['wr']))
      except ZeroDivisionError:
        avg_req_sz = 0

      try:
        avg_queue_sz = (cur[key]['aveq'] - prev[key]['aveq']) / 1000
      except ZeroDivisionError:
        avg_queue_sz = 0

      try:
        avg_wait = ((cur[key]['rd_t'] - prev[key]['rd_t']) +\
                    (cur[key]['wr_t'] - prev[key]['wr_t'])) /\
                   ((cur[key]['rd'] - prev[key]['rd']) +\
                    (cur[key]['wr'] - prev[key]['wr']))
        avg_wait = avg_wait if avg_wait > 0 else 0
      except ZeroDivisionError:
        avg_wait = 0

      try:
        avg_read_wait = (cur[key]['rd_t'] - prev[key]['rd_t']) /\
                        (cur[key]['rd'] - prev[key]['rd'])
      except ZeroDivisionError:
        avg_read_wait = 0

      try:
        avg_write_wait = (cur[key]['wr_t'] - prev[key]['wr_t']) /\
                         (cur[key]['wr'] - prev[key]['wr'])
      except ZeroDivisionError:
        avg_write_wait = 0

      util = self.rateof(prev[key]['use'], cur[key]['use'], delta)

      try:
        svctime = util / self.rateof((prev[key]['rd'] + prev[key]['wr']),
                             (cur[key]['rd'] + cur[key]['wr']), delta)
      except ZeroDivisionError:
        svctime = 0

      data[key] = {
        'reads_per_second': fmt(self.rateof(prev[key]['rd'],
                                cur[key]['rd'], delta)),
        'writes_per_second': fmt(self.rateof(prev[key]['wr'],
                                 cur[key]['wr'], delta)),
        'read_merges_per_second': fmt(self.rateof(prev[key]['rd_m'],
                                      cur[key]['rd_m'], delta)),
        'write_merges_per_second': fmt(self.rateof(prev[key]['wr_m'],
                                       cur[key]['wr_m'], delta)),
        'read_kilobytes_per_second': fmt(self.rateof(prev[key]['rd_s'],
                                         cur[key]['rd_s'], delta) / 2),
        'write_kilobytes_per_second': fmt(self.rateof(prev[key]['wr_s'],
                                          cur[key]['wr_s'], delta) / 2),
        'ios_in_progress': fmt(cur[key]['running'] / delta),
        'average_request_size': fmt(avg_req_sz),
        'average_queue_size': fmt(avg_queue_sz / delta),
        'average_wait': fmt(avg_wait),
        'averate_read_wait': fmt(avg_read_wait),
        'averate_write_wait': fmt(avg_write_wait),
        'svctime': fmt(svctime),
        'util': fmt(util / 10)
      }

    return data



import re
from subprocess import Popen, PIPE

from porkchop.plugin import PorkchopPlugin

def fsgetcache():
  cmd = 'fs getcache'
  data = {}

  # AFS using 38064 of the cache's available 69952 1K byte blocks.
  r1 = re.compile('AFS using (\d+) of.+ (\d+) 1K')

  output = Popen([cmd], stdout=PIPE, shell=True).communicate()[0].splitlines()

  for line in output:
    match = r1.match(line)
    if not match:
      (data['cache_used'], data['cache_total']) = match.groups()

  return data

def fssysname():
  cmd = 'fs sysname'
  data = {}

  output = Popen([cmd], stdout=PIPE, shell=True).communicate()[0].splitlines()

  # 2 possible forms of its output
  # Current sysname list is 'sun4x_510' 'FOO' 'bar'       etc...
  # Current sysname is 'sun4x_510'
  r1 = re.compile('^Current sysname.+is\s+(.+)')

  for line in output:
    match = r1.match(line)
    if not match:
      syses_string = match.group(1).strip()
      syses_string = syses_string.replace("'", "")
      syses_string = syses_string.replace(" ", ",")
      data['sysnames'] = syses_string
        
  return data

class OpenafsPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}

    data = fsgetcache()

    # We know there will be no dict key collisions, so just fold one
    # dict into the other
    data.update(fssysname())

    return data



import socket

from porkchop.plugin import PorkchopPlugin

class RedisPlugin(PorkchopPlugin):
  def _connect(self, host, port):
    try:
      sock = socket.socket()
      sock.connect((host, port))
    except:
      raise

    return sock

  def get_data(self):
    data = self.gendict()

    instance_config = self.config.get('redis', {}).get('instances',
      'localhost:6379')

    instances = [s.strip().split(':') for s in instance_config.split(',')]

    for host, port in instances:
      try:
        sock = self._connect(host, int(port))
        sock.send('info\r\n')

        # first line is the response length in bytes
        resp_hdr = ''
        sock.recv(1)
        while not resp_hdr.endswith('\r\n'):
          resp_hdr += sock.recv(1)

        resp_len = int(resp_hdr.strip())

        resp_data = sock.recv(resp_len)
        sock.send('quit\r\n')
        sock.close()
      except (socket.error, ValueError):
        continue

      for line in resp_data.splitlines():
        k, v = line.split(':', 1)

        # some stat values are CSV, k/v delimited with an '='
        # one of them is allocation_stats but its format is
        # all fucked up
        if ',' in v and k != 'allocation_stats':
          for stat in v.split(','):
            k2, v2 = stat.split('=')
            data[port][k][k2] = v2
        else:
          data[port][k] = v

    return data



import glob
import time

from porkchop.plugin import PorkchopPlugin

def processes_count():
  count = 0

  with open('/proc/stat', 'r') as f:
    for line in f:
      if line.startswith('processes'):
        count = int(line.split()[1])

  return count

def thread_count(f):
  threads = 0

  try:
    with open(f, 'r') as f:
      for line in f:
        if line.startswith('Threads:'):
          count = int(line.split()[1])
  except IOError:
    pass

  return count

class ProcessesPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}

    data['count'] = data['threads'] = 0
    for piddir in glob.glob('/proc/[1-9]*'):
      data['count'] += 1
      data['threads'] += thread_count('%s/status' % piddir)

    if not self.__class__._cache:
      prev = processes_count()
      delta = 1
      time.sleep(delta)
    else:
      prev = self.__class__._cache
      delta = int(time.time() - self.__class__._lastrefresh)

    self.__class__._cache = cur = processes_count()

    data['forkrate'] = self.rateof(prev, cur, delta)

    return data



from porkchop.plugin import PorkchopPlugin

class OpenfilesPlugin(PorkchopPlugin):
  def get_data(self):
    with open('/proc/sys/fs/file-nr', 'r') as f:
      fields = f.read().split()

    data = {
      'count': int(fields[0]) - int(fields[1]),
      'max': fields[2]
    }

    return data



from collections import defaultdict
import glob, os, sys
import socket
import time

import porkchop.plugins
from porkchop.util import PorkchopUtil

class PorkchopPlugin(object):
  config_file = None
  _cache = None
  _data = {}
  _lastrefresh = 0

  def __init__(self):
    self.refresh = 60
    self.force_refresh = False

  @property
  def data(self):
    if self.should_refresh():
      self.config = PorkchopUtil.parse_config(self.config_file)
      self.data = self.get_data()

    return self.__class__._data

  @data.setter
  def data(self, value):
    self.__class__._data = value
    self.__class__._lastrefresh = time.time()
    self.__class__._data['refreshtime'] = int(self.__class__._lastrefresh)
    self.force_refresh = False

  def gendict(self):
    return defaultdict(self.gendict)

  def rateof(self, a, b, ival):
    return (float(b) - float(a)) / ival if (float(b) - float(a)) > 0 else 0

  def should_refresh(self):
    if self.force_refresh:
      return True

    if self.__class__._lastrefresh != 0:
      if time.time() - self.__class__._lastrefresh > self.refresh:
        return True
      else:
        return False
    else:
      return True

  def tcp_socket(self, host, port):
    try:
      sock = socket.socket()
      sock.connect((host, port))
    except socket.error:
      raise

    return sock

  def unix_socket(self, path):
    try:
      sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      sock.connect(path)
    except socket.error:
      raise

    return sock

class PorkchopPluginHandler(object):
  plugins = {}

  def __init__(self, config_dir, directory = None):
    self.config_dir = config_dir
    self.config = PorkchopUtil.parse_config(os.path.join(self.config_dir,
      'porkchop.ini'))

    if directory:
      self.__class__.plugins.update(self.load_plugins(directory))

    self.__class__.plugins.update(self.load_plugins(os.path.dirname(porkchop.plugins.__file__)))

  def load_plugins(self, directory):
    plugins = {}
    sys.path.insert(0, directory)

    try:
      to_load = [p.strip() for p in self.config['porkchop']['plugins'].split(',')]
    except:
      to_load = []

    for infile in glob.glob(os.path.join(directory, '*.py')):
      module_name = os.path.splitext(os.path.split(infile)[1])[0]
      if os.path.basename(infile) != '__init__.py' and \
          (not to_load or module_name in to_load):
        try:
          plugins[module_name] = self.str_to_obj('%s.%sPlugin' % (module_name,
            module_name.capitalize()))
          plugins[module_name].config_file = os.path.join(self.config_dir,
            '%s.ini' % module_name)
        except ImportError:
          pass

    return plugins

  def str_to_obj(self, astr):
    try:
      return globals()[astr]
    except KeyError:
      try:
        __import__(astr)
        mod=sys.modules[astr]
        return mod
      except ImportError:
        module,_,basename=astr.rpartition('.')
        if module:
          mod=self.str_to_obj(module)
          return getattr(mod,basename)
        else:
          raise



from collections import defaultdict
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
import json
from SocketServer import ThreadingMixIn
import sys
import urlparse

from porkchop.plugin import PorkchopPluginHandler

class GetHandler(BaseHTTPRequestHandler):
  def format_output(self, fmt, data):
    if fmt == 'json':
      return json.dumps(data)
    else:
      return '\n'.join(self.json_path(data))

  def json_path(self, data):
    results = []
    def path_helper(data, path, results):
      for key, val in data.iteritems():
        if type(val) in [dict, defaultdict] :
          path_helper(val, '/'.join((path, key)), results)
        else:
          results.append(('%s %s' % (('/'.join((path, key)))\
            .replace('.','_'), val)))

    path_helper(data, '', results)
    return results

  def do_GET(self):
    data = {}
    formats = {'json': 'application/json', 'text': 'text/plain'}
    request = urlparse.urlparse(self.path)

    try:
      (path, fmt) = request.path.split('.')
      if fmt not in formats.keys(): fmt = 'text'
    except ValueError:
      path = request.path
      if self.headers.get('accept', False) == 'application/json':
        fmt = 'json'
      else:
        fmt = 'text'

    if self.headers.get('x-porkchop-refresh', False):
      force_refresh = True
    else:
      force_refresh = False

    module = path.split('/')[1]

    try:
      if module:
        plugin = PorkchopPluginHandler.plugins[module]()
        plugin.force_refresh = force_refresh
        self.log_message('Calling plugin: %s with force=%s' % (module, force_refresh))
        data.update({module: plugin.data})
      else:
        for plugin_name, plugin in PorkchopPluginHandler.plugins.iteritems():
          try:
            plugin.force_refresh = force_refresh
            self.log_message('Calling plugin: %s with force=%s' % (plugin_name, force_refresh))
            # if the plugin has no data,
            # it'll only have one key:
            # refreshtime
            if len(plugin().data) > 1:
              data.update({plugin_name: plugin().data})
          except:
            self.log_error('Error loading plugin: name=%s exception=%s', plugin_name, sys.exc_info())

      if len(data):
        self.send_response(200)
        self.send_header('Content-Type', formats[fmt])
        self.end_headers()
        self.wfile.write(self.format_output(fmt, data) + '\n')
      else:
        raise Exception('Unable to load any plugins')
    except:
      self.log_error('Error: %s', sys.exc_info())
      self.send_response(404)
      self.send_header('Content-Type', formats[fmt])
      self.end_headers()
      self.wfile.write(self.format_output(fmt, {}))

    return

class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
  """ do stuff """



import re
from subprocess import Popen, PIPE

from porkchop.plugin import PorkchopPlugin

class FacterPlugin(PorkchopPlugin):
  def get_data(self):
    d1 = {}
    r1 = re.compile('(.*?)\s=>\s(.*)')

    output = Popen(['facter'], stdout=PIPE, shell=True).communicate()[0]
    for line in output.splitlines():
      match = r1.match(line)
      if match:
        d1[match.group(1)] = match.group(2)

    return d1



import os
import time

from porkchop.plugin import PorkchopPlugin
def read_info():
  data = {}

  with open('/proc/stat', 'r') as f:
    for line in f:
      if line.startswith('cpu'):
        fields = line.split()
        data[fields[0]] = fields[1:]
      else:
        break

  return data

class CpuPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}

    fields = [
      'user',
      'nice',
      'system',
      'idle',
      'iowait',
      'irq',
      'softirq'
    ]

    hz = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

    if not self.__class__._cache:
      prev = read_info()
      delta = 1
      time.sleep(delta)
    else:
      prev = self.__class__._cache
      delta = int(time.time() - self.__class__._lastrefresh)

    self.__class__._cache = cur = read_info()

    for key in cur.keys():
      data.setdefault(key, {})
      for pos in xrange(len(fields)):
        fname = fields[pos]
        data[key].update({
          fname: self.rateof(prev[key][pos], cur[key][pos], delta)
        })

    return data

  def rateof(self, a, b, ival):
    jiffy = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

    return (float(b) - float(a)) / ival * 100 / jiffy





import socket

from porkchop.plugin import PorkchopPlugin

class MemcachedPlugin(PorkchopPlugin):
  def _connect(self, host, port):
    try:
      sock = socket.socket()
      sock.connect((host, port))
    except:
      raise

    return sock

  def get_data(self):
    data = self.gendict()
    resp_data = ''

    instance_config = self.config.get('memcached', {}).get('instances',
      'localhost:11211')

    instances = [s.strip().split(':') for s in instance_config.split(',')]

    for host, port in instances:
      try:
        sock = self._connect(host, int(port))

        sock.send('stats\r\nquit\r\n')

        while not resp_data.endswith('END\r\n'):
          resp_data += sock.recv(1024)

        sock.close()
      except socket.error:
        continue

      for line in resp_data.splitlines():
        if not line.startswith('STAT'): continue
        trash, k, v = line.split()
        data[port][k] = v

    return data



from porkchop.plugin import PorkchopPlugin

class LoadavgPlugin(PorkchopPlugin):
  def get_data(self):
    with open('/proc/loadavg', 'r') as f:
      fields = f.read().strip().split()
      procs = fields[3].split('/')

    return {
      'shortterm': fields[0],
      'midterm': fields[1],
      'longterm': fields[2],
      'processes': {
        'scheduled': procs[0],
        'total': procs[1]
      }
    }



from subprocess import Popen, PIPE
from porkchop.plugin import PorkchopPlugin


class RabbitmqPlugin(PorkchopPlugin):
  def get_vhosts(self):
    return Popen(
      "/usr/sbin/rabbitmqctl list_vhosts",
      stdout=PIPE,
      shell=True
    ).communicate()[0]\
     .strip()\
     .split('\n')[1:-1]

  def get_queues(self, vhost):
    raw_queues = Popen(
      "/usr/sbin/rabbitmqctl list_queues -p %s" % vhost,
      stdout=PIPE,
      shell=True
    ).communicate()[0]\
     .strip()\
     .split('\n')[1:-1]

    queues = {}
    for line in raw_queues:
      try:
        queue, size = line.rsplit('\t', 1)
      except:
        pass
      else:
        queues[queue] = int(size)

    return queues

  def get_data(self):
    output = {}

    vhosts = self.get_vhosts()
    for vhost in vhosts:
      queues = self.get_queues(vhost)
      if vhost == '/':
        vhost = 'default'
      if queues:
        output[vhost] = queues

    return {'vhost': output}



import json
import urllib2
from porkchop.plugin import PorkchopPlugin

class RiakPlugin(PorkchopPlugin):
  def get_data(self):
    stats_url = self.config.get('riak', {}).get('stats_url',
      'http://localhost:8098/stats')

    stats = json.loads(urllib2.urlopen(stats_url).read())

    for key in stats.keys():
      if type(stats[key]) == list:
        stats[key] = ', '.join(stats[key])

    return stats



from porkchop.plugin import PorkchopPlugin

class MeminfoPlugin(PorkchopPlugin):
  def get_data(self):
    data = {}
    with open('/proc/meminfo', 'r') as f:
      for line in f:
        lhs, rhs = line.split(':')
        lhs = lhs.replace('(', '_').replace(')', '').lower()
        data[lhs] = rhs.strip().split()[0]

      data['memused'] = int(data['memtotal']) - \
                     (int(data['memfree']) + \
                      int(data['buffers']) + \
                      int(data['cached']))
      data['swapused'] = int(data['swaptotal']) - int(data['swapfree'])

    return data



#!/usr/bin/env python

import glob
from setuptools import setup, find_packages

version = '0.7'

setup(name='porkchop',
  version=version,
  description='Porkchop is a simple HTTP-based system information server',
  long_description='Porkchop is a simple HTTP-based system information server',
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Topic :: System :: Networking :: Monitoring'
  ],
  keywords='',
  author='Scott Smith',
  author_email='scott@disqus.com',
  url='http://github.com/disqus/porkchop',
  license='Apache 2.0',
  packages=find_packages(),
  include_package_data=True,
  zip_safe=False,
  install_requires=['requests'],
  entry_points={
    'console_scripts': [
      'porkchop = porkchop.commandline:main',
      'porkchop-collector = porkchop.commandline:collector'
    ],
  }
)



from collections import defaultdict
from subprocess import Popen, PIPE

from porkchop.plugin import PorkchopPlugin

class TcpudpPlugin(PorkchopPlugin):
  def get_data(self):
    command = 'ss -tuna'
    data = self.gendict()

    output = Popen([command], stdout=PIPE, shell=True)\
                   .communicate()[0]\
                   .splitlines()

    output.pop(0)

    for line in output:
      fields = line.split()
      proto = fields[0]

      if proto == 'tcp':
        state = fields[1].lower()
        data[proto][state].setdefault('value', 0)
        data[proto][state]['value'] += 1
      else:
        data[proto].setdefault('value', 0)
        data[proto]['value'] += 1

    return dict(data)



import requests

from porkchop.plugin import PorkchopPlugin

class ApachePlugin(PorkchopPlugin):
  def parse_scoreboard(self, scoreboard):
    sb = {}
    scoreboard_keys = {
      '_': 'waiting',
      'S': 'starting',
      'R': 'reading',
      'W': 'sending',
      'K': 'keepalive',
      'D': 'dnslookup',
      'C': 'closing',
      'L': 'logging',
      'G': 'finishing',
      'I': 'idle_cleanup',
      '.': 'open'
    }

    for chr in list(scoreboard):
      sb.setdefault(scoreboard_keys[chr], 0)
      sb[scoreboard_keys[chr]] += 1

    return sb

  def get_data(self):
    data = {}

    r = requests.get('http://localhost/server-status?auto')
    for line in r.content.strip('\n').splitlines():
      (key, val) = line.split(':', 1)

      if key == 'Scoreboard':
        data['scoreboard'] = self.parse_scoreboard(val.strip())
      else:
        data[key.replace(' ', '_').lower()] = val.strip()

    return data
